from dune.codegen.generation import (backend,
                                     kernel_cached,
                                     get_backend,
                                     instruction,
                                     temporary_variable,
                                     globalarg,
                                     class_member,
                                     initializer_list,
                                     include_file,)
from dune.codegen.tools import get_pymbolic_basename
from dune.codegen.loopy.target import type_floatingpoint
from dune.codegen.pdelab.basis import (declare_cache_temporary,
                                       name_localbasis_cache,
                                       type_localbasis,
                                       FEM_name_mangling)
from dune.codegen.pdelab.driver import (isPk,
                                        isQk,
                                        )
from dune.codegen.pdelab.geometry import world_dimension
from dune.codegen.pdelab.quadrature import pymbolic_quadrature_position_in_cell
from dune.codegen.pdelab.spaces import type_leaf_gfs
from dune.codegen.pdelab.restriction import restricted_name
from dune.codegen.blockstructured.spaces import lfs_inames
from dune.codegen.blockstructured.tools import tensor_index_to_sequential_index

from ufl import MixedElement

import pymbolic.primitives as prim


# define FE basis explicitly in localoperator
@backend(interface="typedef_localbasis", name="blockstructured")
@class_member(classtag="operator")
def typedef_localbasis(element, name):
    df = "typename {}::Traits::GridView::ctype".format(type_leaf_gfs(element))
    r = type_floatingpoint()
    dim = world_dimension()
    if isPk(element):
        if dim == 1:
            include_file("dune/localfunctions/lagrange/pk1d/pk1dlocalbasis.hh", filetag="operatorfile")
            basis_type = "Pk1DLocalBasis<{}, {}, {}>".format(df, r, element._degree)
        elif dim == 2:
            include_file("dune/localfunctions/lagrange/pk2d/pk2dlocalbasis.hh", filetag="operatorfile")
            basis_type = "Pk2DLocalBasis<{}, {}, {}>".format(df, r, element._degree)
        elif dim == 3:
            include_file("dune/localfunctions/lagrange/pk3d/pk3dlocalbasis.hh", filetag="operatorfile")
            basis_type = "Pk3DLocalBasis<{}, {}, {}>".format(df, r, element._degree)
        else:
            raise NotImplementedError("P{} in {}D is not implemented".format(element._degree, dim))
    elif isQk(element):
        include_file("dune/localfunctions/lagrange/qk/qklocalbasis.hh", filetag="operatorfile")
        basis_type = "QkLocalBasis<{}, {}, {}, {}>".format(df, r, element._degree, dim)
    else:
        raise NotImplementedError("Element type not known in code generation")
    return "using {} = Dune::{};".format(name, basis_type)


@class_member(classtag="operator")
def define_localbasis(leaf_element, name):
    localBasis_type = type_localbasis(leaf_element)
    initializer_list(name, (), classtag="operator")
    return "const {} {};".format(localBasis_type, name)


def name_localbasis(leaf_element):
    name = "{}_microElementBasis".format(FEM_name_mangling(leaf_element))
    globalarg(name)
    define_localbasis(leaf_element, name)
    return name


@kernel_cached
def evaluate_basis(leaf_element, name, restriction):
    temporary_variable(name, shape=((leaf_element.degree() + 1)**world_dimension(), 1),
                       decl_method=declare_cache_temporary(leaf_element, restriction, 'Function'))
    cache = name_localbasis_cache(leaf_element)
    qp = pymbolic_quadrature_position_in_cell(restriction)
    localbasis = name_localbasis(leaf_element)
    instruction(inames=get_backend("quad_inames")(),
                code='{} = {}.evaluateFunction({}, {});'.format(name, cache, str(qp), localbasis),
                assignees=frozenset({name}),
                read_variables=frozenset({get_pymbolic_basename(qp)}),
                )


def pymbolic_basis(leaf_element, restriction, number, context=''):
    assert not isinstance(leaf_element, MixedElement)
    name = "phi_{}".format(FEM_name_mangling(leaf_element))
    name = restricted_name(name, restriction)
    evaluate_basis(leaf_element, name, restriction)
    inames = lfs_inames(leaf_element, restriction, number, context=context)

    return prim.Subscript(prim.Variable(name), (tensor_index_to_sequential_index(inames, leaf_element.degree() + 1), 0))


@kernel_cached
def evaluate_reference_gradient(leaf_element, name, restriction):
    temporary_variable(name, shape=((leaf_element.degree() + 1)**world_dimension(), 1, world_dimension()),
                       decl_method=declare_cache_temporary(leaf_element, restriction, 'Jacobian'))
    cache = name_localbasis_cache(leaf_element)
    qp = pymbolic_quadrature_position_in_cell(restriction)
    localbasis = name_localbasis(leaf_element)
    instruction(inames=get_backend("quad_inames")(),
                code='{} = {}.evaluateJacobian({}, {});'.format(name, cache, str(qp), localbasis),
                assignees=frozenset({name}),
                read_variables=frozenset({get_pymbolic_basename(qp)}),
                )


def pymbolic_reference_gradient(leaf_element, restriction, number, context=''):
    assert not isinstance(leaf_element, MixedElement)
    name = "js_{}".format(FEM_name_mangling(leaf_element))
    name = restricted_name(name, restriction)
    evaluate_reference_gradient(leaf_element, name, restriction)
    inames = lfs_inames(leaf_element, restriction, number, context=context)

    return prim.Subscript(prim.Variable(name), (tensor_index_to_sequential_index(inames, leaf_element.degree() + 1), 0))
