import loopy as lp
import numpy as np
import pymbolic.primitives as prim

from loopy.match import Tagged, Id, Writes, Or

from dune.codegen.generation import get_global_context_value
from dune.codegen.loopy.target import dtype_floatingpoint
from dune.codegen.loopy.temporary import DuneTemporaryVariable
from dune.codegen.loopy.symbolic import substitute
from dune.codegen.loopy.vcl import get_vcl_type_size, VCLPermute, VCLLoad, VCLStore
from dune.codegen.options import get_form_option
from dune.codegen.pdelab.geometry import world_dimension
from dune.codegen.tools import get_pymbolic_basename


def add_vcl_temporaries(knl):
    vector_alias = [a for a in knl.arg_dict if a.endswith('alias')]

    # add new temporaries for vectors
    # hope one read insn doesn't have two different reads from the same temporary
    new_vec_temporaries = dict()
    new_insns = []
    init_iname = 'init_vec'
    from islpy import BasicSet
    init_domain = BasicSet("{{ [{0}] : 0<={0}<{1} }}".format(init_iname, get_vcl_type_size(dtype_floatingpoint())))
    for alias in vector_alias:
        vector_name = alias.replace('alias', 'vec')
        new_vec_temporaries[vector_name] = DuneTemporaryVariable(vector_name, dtype=np.float64,
                                                                 shape=(get_vcl_type_size(np.float64),), managed=True,
                                                                 scope=lp.temp_var_scope.PRIVATE, dim_tags=('vec',))
        # write once to the vector such that loopy won't complain
        new_insns.append(lp.Assignment(assignee=prim.Subscript(prim.Variable(vector_name), prim.Variable(init_iname)),
                                       expression=0, within_inames=frozenset({init_iname}),
                                       id='init_' + vector_name))

    from loopy.kernel.data import VectorizeTag
    return knl.copy(instructions=knl.instructions + new_insns, domains=knl.domains + [init_domain],
                    temporary_variables=dict(**knl.temporary_variables, **new_vec_temporaries),
                    iname_to_tag=dict(**knl.iname_to_tag, **{init_iname: VectorizeTag()}))


def add_vcl_accum_insns(knl, iname_inner, iname_outer):
    nptype = dtype_floatingpoint()
    vcl_size = get_vcl_type_size(np.float64)

    from loopy.match import Tagged
    accum_insns = set(lp.find_instructions(knl, Tagged('accum')))

    new_insns = []
    vng = knl.get_var_name_generator()
    idg = knl.get_instruction_id_generator()
    new_vec_temporaries = dict()
    for insn in knl.instructions:
        # somehow CInstructions are not hashable....
        if isinstance(insn, lp.MultiAssignmentBase) and insn in accum_insns:
            # write accum expr as "r = expr + r"
            expr_without_r = prim.Sum(tuple(e for e in insn.expression.children if not e == insn.assignee))

            inames_micro = set((i for i in insn.within_inames if i.startswith('micro')))
            iname_ix = next((i for i in inames_micro if i.endswith("_x")))

            # need inames for head and tail handling a priori
            from loopy.match import Not, All
            replace_head_inames = dict()
            replace_tail_inames = dict()
            for iname in inames_micro - frozenset({iname_ix}):
                head = iname + '_head'
                tail = iname + '_tail'
                replace_head_inames[iname] = prim.Variable(head)
                replace_tail_inames[iname] = prim.Variable(tail)
                knl = lp.duplicate_inames(knl, iname, Not(All()), suffix='_tail')
                knl = lp.duplicate_inames(knl, iname, Not(All()), suffix='_head')
            inames_head = frozenset((var.name for var in replace_head_inames.values()))
            inames_tail = frozenset((var.name for var in replace_tail_inames.values()))

            # erstelle a[iy] und b
            identifier_left = vng('left_node')
            identifier_right = vng('right_node')
            new_vec_temporaries[identifier_left] = DuneTemporaryVariable(identifier_left, dtype=np.float64,
                                                                         shape=(2,) * (world_dimension() - 1) +
                                                                               (vcl_size,),
                                                                         managed=True, scope=lp.temp_var_scope.PRIVATE,
                                                                         dim_tags=('f',) * (world_dimension() - 1) +
                                                                                  ('vec',))
            new_vec_temporaries[identifier_right] = DuneTemporaryVariable(identifier_right, dtype=np.float64,
                                                                          shape=(vcl_size,), managed=True,
                                                                          scope=lp.temp_var_scope.PRIVATE, dim_tags=('vec',))

            var_left = prim.Subscript(prim.Variable(identifier_left),
                                      tuple(prim.Variable(i) for i in sorted(inames_micro - frozenset({iname_ix}))) +
                                      (prim.Variable(iname_inner),))
            var_right = prim.Subscript(prim.Variable(identifier_right), (prim.Variable(iname_inner),))

            # init a
            id_init_a = idg('insn_init_' + identifier_left)
            new_insns.append(lp.Assignment(assignee=substitute(var_left, replace_head_inames),
                                           expression=0,
                                           id=id_init_a,
                                           within_inames=(insn.within_inames - frozenset({iname_outer}) -
                                                          inames_micro) | inames_head,
                                           tags=frozenset({'head'})))

            # setze werte für a und b
            expr_right = substitute(expr_without_r, {iname_ix: 1})
            expr_left = prim.Sum((substitute(expr_without_r, {iname_ix: 0}), var_left))

            id_set_left = idg('insn_' + identifier_left)
            id_set_right = idg('insn_' + identifier_right)
            new_insns.append(lp.Assignment(assignee=var_right,
                                           expression=expr_right,
                                           id=id_set_right,
                                           depends_on=insn.depends_on,
                                           within_inames=insn.within_inames - frozenset({iname_ix})))
            new_insns.append(lp.Assignment(assignee=var_left,
                                           expression=expr_left,
                                           id=id_set_left,
                                           depends_on=insn.depends_on | frozenset({id_init_a}),
                                           within_inames=insn.within_inames - frozenset({iname_ix})))

            # r+=a[iy]
            id_accum = idg('insn_mod_accum')
            expr_accum = prim.Sum((var_left,
                                   prim.Call(VCLPermute(nptype, vcl_size, (-1,) + tuple(range(vcl_size - 1))),
                                             (var_right,)),
                                   substitute(insn.assignee, {iname_ix: 0})))
            new_insns.append(lp.Assignment(assignee=substitute(insn.assignee, {iname_ix: 0}),
                                           expression=expr_accum,
                                           id=id_accum,
                                           depends_on=insn.depends_on | frozenset({id_set_left,
                                                                                   id_init_a, id_set_right}),
                                           within_inames=insn.within_inames - frozenset({iname_ix}),
                                           tags=frozenset({'accum'})))
            # a[iy] = permute
            id_permute = idg('insn_permute')
            expr_permute = prim.Call(VCLPermute(nptype, vcl_size, (vcl_size - 1,) + (-1,) * (vcl_size - 1)),
                                     (var_right,))
            new_insns.append(lp.Assignment(assignee=var_left,
                                           expression=expr_permute,
                                           id=id_permute,
                                           depends_on=insn.depends_on | frozenset({id_set_left, id_init_a, id_set_right,
                                                                                   id_accum}),
                                           within_inames=insn.within_inames - frozenset({iname_ix})
                                           ))

            # tail handling, uses tail alias
            id_accum_tail = idg('insn_accum_tail')
            subst_map = {iname_inner: vcl_size - 1, iname_outer: get_form_option("number_of_blocks") // vcl_size - 1,
                         iname_ix: 1, insn.assignee_name: prim.Variable(insn.assignee_name + '_tail'),
                         **replace_tail_inames}
            assignee_tail = substitute(insn.assignee, subst_map)
            expr_tail = prim.Sum((substitute(var_left, {iname_inner: 0, **replace_tail_inames}), assignee_tail))

            write_to_tail_ids = tuple(i.id for i in lp.find_instructions(knl,
                                                                         Writes(get_pymbolic_basename(assignee_tail))))

            new_insns.append(lp.Assignment(assignee=assignee_tail,
                                           expression=expr_tail,
                                           id=id_accum_tail,
                                           depends_on=(frozenset({id_accum, id_permute, id_set_left, id_init_a}) |
                                                       frozenset(write_to_tail_ids)),
                                           within_inames=(insn.within_inames - frozenset({iname_inner, iname_outer}) -
                                                          inames_micro) | inames_tail,
                                           tags=frozenset({'tail'})))
        else:
            new_insns.append(insn)

    return knl.copy(instructions=new_insns,
                    temporary_variables=dict(**knl.temporary_variables, **new_vec_temporaries))


def add_vcl_access(knl, iname_inner):
    from loopy.match import Reads, Tagged
    accum_insns = set((insn.id for insn in lp.find_instructions(knl, Tagged('accum'))))
    read_insns = set((insn.id for insn in lp.find_instructions(knl, Reads('*alias'))))
    vectorized_insns = accum_insns | read_insns

    from loopy.symbolic import CombineMapper
    from loopy.symbolic import IdentityMapper

    class AliasIndexCollector(CombineMapper, IdentityMapper):
        def combine(self, values):
            return sum(values, tuple())

        def map_constant(self, expr):
            return tuple()

        map_variable = map_constant
        map_function_symbol = map_constant
        map_loopy_function_identifier = map_constant

        def map_subscript(self, expr):
            if expr.aggregate.name.endswith('alias'):
                return expr.aggregate, expr.index_tuple
            else:
                return tuple()

    # add load instructions if the vector is read, based on the read instruction
    # TODO brauche mehrere vectoren, falls in einer insn von einem alias mit unterschiedlichen idx gelesen wird
    # muss dafür eigentlich nur den namen des vectors anpassen
    idg = knl.get_instruction_id_generator()
    aic = AliasIndexCollector()
    load_insns = []
    read_dependencies = dict()
    vectorized_insn_to_vector_names = dict()
    for id in read_insns:
        insn = knl.id_to_insn[id]

        alias, index = aic(insn.expression)
        name_alias = alias.name
        name_vec = name_alias.replace('alias', 'vec')
        vectorized_insn_to_vector_names[id] = (name_alias, name_vec)

        # compute index without vec iname
        strides = tuple(tag.stride for tag in knl.arg_dict[name_alias].dim_tags)
        index = prim.Sum(tuple(prim.Product((i, s)) for i, s in zip(index, strides)
                               if i != 0 and i.name != iname_inner))

        # find write insns
        write_ids = frozenset(i.id for i in lp.find_instructions(knl, Or((Writes(name_vec), Writes(name_vec)))))

        # add load instruction
        load_id = idg('insn_' + name_vec + '_load')
        call_load = prim.Call(VCLLoad(name_vec), (prim.Sum((prim.Variable(name_alias), index)),))
        load_insns.append(lp.CallInstruction(assignees=(), expression=call_load,
                                             id=load_id, within_inames=insn.within_inames | insn.reduction_inames(),
                                             depends_on=insn.depends_on | write_ids,))
        read_dependencies.setdefault(id, set())
        read_dependencies[id].add(load_id)

    # add store instructions if the vector is written, based on the write instruction
    store_insns = []
    for id in accum_insns:
        insn = knl.id_to_insn[id]

        alias, index = aic(insn.expression)
        name_alias = alias.name
        name_vec = name_alias.replace('alias', 'vec')
        vectorized_insn_to_vector_names[id] = (name_alias, name_vec)

        # flat index without vec iname
        strides = tuple(tag.stride for tag in knl.arg_dict[name_alias].dim_tags)
        index = prim.Sum(tuple(prim.Product((i, s)) for i, s in zip(index, strides)
                               if i != 0 and i.name != iname_inner))

        # find write insns
        write_ids = frozenset(i.id for i in lp.find_instructions(knl, Or((Writes(name_vec), Writes(name_vec)))))

        # add store instruction
        store_id = idg('insn_' + name_vec + '_store')
        call_store = prim.Call(VCLStore(name_vec), (prim.Sum((prim.Variable(name_alias), index)),))
        store_insns.append(lp.CallInstruction(assignees=(), expression=call_store,
                                              id=store_id, within_inames=insn.within_inames,
                                              depends_on=(insn.depends_on | frozenset({id}) | read_dependencies[id] |
                                                          write_ids)))

    # replace alias with vcl vector, except for accumulation assignee
    vector_alias = [a for a in knl.arg_dict if a.endswith('alias')]
    dim = world_dimension()
    dim_names = ["x", "y", "z"] + [str(i) for i in range(4, dim + 1)]
    # remove CInstructions since loopy extract expects to get only assignments
    knl_without_cinsn = knl.copy(instructions=[insn for insn in knl.instructions
                                               if not isinstance(insn, lp.CInstruction)])
    for alias in vector_alias:
        # Rename lhs which would match the substitution rule since loopy doesn't want substitutions as lhs
        new_insns = []
        for insn in knl_without_cinsn.instructions:
            if isinstance(insn, lp.Assignment) and isinstance(insn.assignee, prim.Subscript):
                if insn.assignee.aggregate.name == alias:
                    new_insns.append(insn.copy(assignee=prim.Subscript(prim.Variable('dummy_' + alias),
                                                                       insn.assignee.index_tuple)))
                    pass
                else:
                    new_insns.append(insn)
            else:
                new_insns.append(insn)
        knl_without_cinsn = knl_without_cinsn.copy(instructions=new_insns)

        # substitution rule for alias[ex_outer,ex_inner, ey, ix, iy] -> vec[ex_inner]
        parameters = 'ex_o,ex_i,' + ','.join(['e' + d for d in dim_names[1:dim]]) + \
                     ',ix,' + ','.join(['i' + d for d in dim_names[1:dim]])
        knl_without_cinsn = lp.extract_subst(knl_without_cinsn, alias + '_subst', '{}[{}]'.format(alias, parameters),
                                             parameters=parameters)
        new_subst = knl_without_cinsn.substitutions.copy()
        rule = new_subst[alias + '_subst']
        rule.expression = prim.Subscript(prim.Variable(alias.replace('alias', 'vec')), (prim.Variable('ex_i'),))
        knl_without_cinsn = knl_without_cinsn.copy(substitutions=new_subst)

    from loopy.match import All
    knl_without_cinsn = lp.expand_subst(knl_without_cinsn, All())
    knl = knl_without_cinsn.copy(instructions=knl_without_cinsn.instructions + [insn for insn in knl.instructions
                                                                                if isinstance(insn, lp.CInstruction)])

    # add store and load dependencies and set right accumulation assignee
    new_insns = []
    for insn in knl.instructions:
        if insn.id not in vectorized_insns:
            new_insns.append(insn)
        else:
            # find write insns
            name_alias, name_vec = vectorized_insn_to_vector_names[insn.id]
            write_ids = frozenset(i.id for i in lp.find_instructions(knl, Or((Writes(name_vec), Writes(name_vec)))))
            if insn.id in accum_insns:
                assignee_alias = insn.assignee
                try:
                    assignee_vec = next((expr for expr in insn.expression.children
                                         if isinstance(expr, prim.Subscript) and
                                         expr.aggregate.name.replace('vec', 'alias') ==
                                         assignee_alias.aggregate.name.replace('dummy_', '')))
                except StopIteration:
                    from dune.codegen.error import CodegenVectorizationError
                    raise CodegenVectorizationError
                new_insns.append(insn.copy(assignee=assignee_vec,
                                           depends_on=(insn.depends_on | read_dependencies[insn.id] |
                                                       write_ids)))
            else:
                new_insns.append(insn.copy(depends_on=(insn.depends_on | read_dependencies[insn.id] |
                                                       write_ids)))

    return knl.copy(instructions=new_insns + load_insns + store_insns)


def find_accumulation_inames(knl):
    inames = set()
    for insn in knl.instructions:
        if any((n.startswith('r_') and n.endswith('alias') for n in insn.write_dependency_names())):
            inames |= insn.within_inames

    inames = set((i for i in inames if i.startswith('micro') and not i.endswith('_x')))

    return inames


def add_iname_array(knl, vec_iname):
    insns_with_macro_points = lp.find_instructions(knl, Tagged(vec_iname))

    if insns_with_macro_points:
        array_name = vec_iname + '_arr'
        vector_name = vec_iname + '_vec'

        new_temporaries = dict()
        new_temporaries[array_name] = DuneTemporaryVariable(array_name, managed=True,
                                                            shape=(get_form_option('number_of_blocks'),),
                                                            scope=lp.temp_var_scope.PRIVATE, dtype=np.float64,
                                                            base_storage=vec_iname + '_buff',
                                                            _base_storage_access_may_be_aliasing=True)
        new_temporaries[vector_name] = DuneTemporaryVariable(vector_name, managed=True,
                                                             shape=(get_form_option('number_of_blocks'),),
                                                             scope=lp.temp_var_scope.PRIVATE, dtype=np.float64,
                                                             base_storage=vec_iname + '_buff',
                                                             _base_storage_access_may_be_aliasing=True)
        silenced_warning = ["read_no_write({})".format(vector_name)]

        new_insns = [lp.Assignment(assignee=prim.Subscript(prim.Variable(array_name), (prim.Variable(vec_iname),)),
                                   expression=prim.Variable(vec_iname),
                                   id='init_{}_buffer'.format(vec_iname),
                                   within_inames=frozenset({vec_iname}), within_inames_is_final=True)]

        replacemap = dict()
        replacemap[vec_iname] = prim.Subscript(prim.Variable(vector_name), (prim.Variable(vec_iname),))

        for insn in knl.instructions:
            if insn in insns_with_macro_points:
                transformed_insn = insn.with_transformed_expressions(lambda expr: substitute(expr, replacemap))
                new_insns.append(transformed_insn.copy(depends_on='init_{}_buffer'.format(vec_iname)))
            else:
                new_insns.append(insn)

        knl = knl.copy(instructions=new_insns,
                       temporary_variables=dict(**knl.temporary_variables, **new_temporaries),
                       silenced_warnings=knl.silenced_warnings + silenced_warning)

        knl = lp.duplicate_inames(knl, [vec_iname], within=Id('init_{}_buffer'.format(vec_iname)))

    return knl


def vectorize_micro_elements(knl):
    vec_iname = "subel_x"
    if vec_iname in knl.all_inames() and get_global_context_value('integral_type') == 'cell':
        vcl_size = get_vcl_type_size(np.float64)

        assert get_form_option('number_of_blocks') % vcl_size == 0

        knl = add_iname_array(knl, vec_iname)

        knl = lp.split_iname(knl, vec_iname, vcl_size, inner_tag='vec')
        array_alias = [a for a in knl.arg_dict.keys() if a.endswith('alias') or a.endswith('tail')]
        iname_vector = [a for a in knl.temporary_variables.keys() if a.endswith('vec')]
        knl = lp.split_array_axis(knl, array_alias + iname_vector, 0, vcl_size)
        knl = lp.tag_array_axes(knl, iname_vector, ('c', 'vec'))

        knl = add_vcl_temporaries(knl)
        knl = add_vcl_accum_insns(knl, vec_iname + '_inner', vec_iname + '_outer')
        knl = add_vcl_access(knl, vec_iname + '_inner')
    return knl
