""" The pdelab specific parts of the code generation process """

# Trigger some imports that are needed to have all backend implementations visible
# to the selection mechanisms
from dune.codegen.generation import (get_backend)
from dune.codegen.options import option_switch
from dune.codegen.pdelab.argument import (pymbolic_apply_function,
                                          pymbolic_apply_function_gradient,
                                          pymbolic_trialfunction,
                                          pymbolic_trialfunction_gradient,
                                          pymbolic_coefficient
                                          )
from dune.codegen.pdelab.basis import (pymbolic_basis,
                                       pymbolic_reference_gradient,
                                       )
from dune.codegen.pdelab.function import pymbolic_gridfunction
from dune.codegen.pdelab.geometry import (component_iname,
                                          pymbolic_cell_volume,
                                          pymbolic_facet_area,
                                          pymbolic_facet_jacobian_determinant,
                                          pymbolic_jacobian_determinant,
                                          pymbolic_jacobian_inverse,
                                          pymbolic_unit_inner_normal,
                                          pymbolic_unit_outer_normal,
                                          to_global,
                                          )
from dune.codegen.pdelab.index import (name_index,
                                       )
from dune.codegen.pdelab.quadrature import (pymbolic_quadrature_weight,
                                            pymbolic_quadrature_position,
                                            quadrature_inames,
                                            )
from dune.codegen.pdelab.spaces import (lfs_inames,
                                        )
from dune.codegen.pdelab.tensors import (pymbolic_list_tensor,
                                         pymbolic_identity,
                                         pymbolic_matrix_inverse,
                                         )


class PDELabInterface(object):
    def __init__(self):
        # The visitor instance will be registered by its init method
        self.visitor = None

    # Accumulation interfaces
    def get_accumulation_info(self, expr, visitor):
        from dune.codegen.pdelab.localoperator import get_accumulation_info
        return get_accumulation_info(expr, visitor)

    def list_accumulation_infos(self, expr, visitor):
        from dune.codegen.pdelab.localoperator import list_accumulation_infos
        return list_accumulation_infos(expr, visitor)

    def generate_accumulation_instruction(self, expr, visitor):
        from dune.codegen.pdelab.localoperator import generate_accumulation_instruction
        return generate_accumulation_instruction(expr, visitor)

    #
    # TODO: The following ones are actually entirely PDELab independent!
    # They should be placed elsewhere and be used directly in the visitor.
    #

    def component_iname(self, context=None, count=None):
        return component_iname(context=context, count=count)

    def name_index(self, ind):
        return name_index(ind)

    #
    # Local function space related generator functions
    #

    def lfs_inames(self, element, restriction, number=None, context=''):
        return lfs_inames(element, restriction, number, context)

    def initialize_function_spaces(self, expr, visitor):
        from dune.codegen.pdelab.spaces import initialize_function_spaces
        return initialize_function_spaces(expr, visitor)

    #
    # Test and trial function related generator functions
    #

    # TODO: should pymbolic_basis also pass the interface?
    def pymbolic_basis(self, element, restriction, number, context=''):
        return pymbolic_basis(element, restriction, number, context)

    # TODO: should pymbolic_reference_gradient also pass the interface?
    def pymbolic_reference_gradient(self, element, restriction, number, context=''):
        return pymbolic_reference_gradient(element, restriction, number, context)

    def pymbolic_trialfunction_gradient(self, element, restriction, index):
        return pymbolic_trialfunction_gradient(self.visitor, element, restriction, index)

    def pymbolic_apply_function_gradient(self, element, restriction, index):
        return pymbolic_apply_function_gradient(self.visitor, element, restriction, index)

    def pymbolic_trialfunction(self, element, restriction, index):
        return pymbolic_trialfunction(self.visitor, element, restriction, index)

    def pymbolic_apply_function(self, element, restriction, index):
        return pymbolic_apply_function(self.visitor, element, restriction, index)

    def pymbolic_gridfunction(self, coeff, restriction, grad):
        return pymbolic_gridfunction(coeff, restriction, grad)

    #
    # Tensor expression related generator functions
    #

    def pymbolic_list_tensor(self, o):
        return pymbolic_list_tensor(o, self.visitor)

    def pymbolic_identity(self, o):
        return pymbolic_identity(o)

    def pymbolic_matrix_inverse(self, o):
        return pymbolic_matrix_inverse(o, self.visitor)

    #
    # Geometry related generator functions
    #

    def pymbolic_spatial_coordinate(self):
        return to_global(pymbolic_quadrature_position())

    def pymbolic_facet_jacobian_determinant(self):
        return pymbolic_facet_jacobian_determinant()

    def pymbolic_jacobian_determinant(self):
        return pymbolic_jacobian_determinant()

    def pymbolic_jacobian_inverse(self, i, j, restriction):
        return pymbolic_jacobian_inverse(i, j, restriction)

    def pymbolic_unit_inner_normal(self):
        return pymbolic_unit_inner_normal()

    def pymbolic_unit_outer_normal(self):
        return pymbolic_unit_outer_normal()

    def pymbolic_cell_volume(self, restriction):
        return pymbolic_cell_volume(restriction)

    def pymbolic_facet_area(self):
        return pymbolic_facet_area()

    #
    # Quadrature related generator functions
    #

    def pymbolic_quadrature_weight(self):
        return pymbolic_quadrature_weight()

    # TODO Should this be part of interface or not?
    def quadrature_inames(self):
        return quadrature_inames()
