""" Generator functions related to any input and output of accumulation kernels

Namely:
* Coefficient containers
* accumulation object (r, jac...)
"""

from dune.codegen.generation import (domain,
                                     function_mangler,
                                     iname,
                                     valuearg,
                                     get_global_context_value,
                                     kernel_cached,
                                     backend
                                     )
from dune.codegen.loopy.target import dtype_floatingpoint
from dune.codegen.pdelab.index import name_index
from dune.codegen.pdelab.basis import (evaluate_coefficient,
                                       evaluate_coefficient_gradient,
                                       )
from dune.codegen.pdelab.spaces import (lfs_iname,
                                        name_lfs_bound,
                                        )
from dune.codegen.pdelab.restriction import restricted_name
from dune.codegen.ufl.modified_terminals import Restriction

from pymbolic.primitives import Call, Subscript, Variable

from loopy import CallMangleInfo
from loopy.symbolic import FunctionIdentifier
from loopy.types import NumpyType

import numpy


class CoefficientAccess(FunctionIdentifier):
    def __init__(self, container):
        self.container = container

    def __getinitargs__(self):
        return (self.container,)

    @property
    def name(self):
        return self.container


@function_mangler
def coefficient_mangler(target, func, dtypes):
    if isinstance(func, CoefficientAccess):
        return CallMangleInfo(func.name, (NumpyType(dtype_floatingpoint()),), (NumpyType(str), NumpyType(numpy.int32)))


class PDELabAccumulationFunction(FunctionIdentifier):
    def __init__(self, accumobj, rank):
        self.accumobj = accumobj
        self.rank = rank

        assert rank in (1, 2)

    def __getinitargs__(self):
        return (self.accumobj, self.rank)

    @property
    def name(self):
        return '{}.accumulate'.format(self.accumobj)


@function_mangler
def accumulation_mangler(target, func, dtypes):
    if isinstance(func, PDELabAccumulationFunction):
        if func.rank == 1:
            return CallMangleInfo(func.name,
                                  (),
                                  (NumpyType(str),
                                   NumpyType(numpy.int32),
                                   NumpyType(dtype_floatingpoint()),
                                   )
                                  )
        if func.rank == 2:
            return CallMangleInfo(func.name,
                                  (),
                                  (NumpyType(str),
                                   NumpyType(numpy.int32),
                                   NumpyType(str),
                                   NumpyType(numpy.int32),
                                   NumpyType(dtype_floatingpoint()),
                                   )
                                  )


def pymbolic_trialfunction_gradient(visitor, element, restriction, index):
    rawname = "gradu_{}".format(index)
    name = restricted_name(rawname, restriction)
    container = name_coefficientcontainer(restriction)
    evaluate_coefficient_gradient(visitor, element, name, container, restriction, index)
    return Variable(name)


def pymbolic_trialfunction(visitor, element, restriction, index):
    rawname = "u_{}".format(index)
    name = restricted_name(rawname, restriction)
    container = name_coefficientcontainer(restriction)
    evaluate_coefficient(visitor, element, name, container, restriction, index)
    return Variable(name)


def pymbolic_apply_function_gradient(visitor, element, restriction, index):
    rawname = "gradz_func_{}".format(index)
    name = restricted_name(rawname, restriction)
    container = name_applycontainer(restriction)
    evaluate_coefficient_gradient(visitor, element, name, container, restriction, index)
    return Variable(name)


def pymbolic_apply_function(visitor, element, restriction, index):
    rawname = "z_func_{}".format(index)
    name = restricted_name(rawname, restriction)
    container = name_applycontainer(restriction)
    evaluate_coefficient(visitor, element, name, container, restriction, index)
    return Variable(name)


def name_coefficientcontainer(restriction):
    name = restricted_name("x", restriction)
    return name


def name_applycontainer(restriction):
    name = restricted_name("z", restriction)
    return name


@kernel_cached
@backend(interface="pymbolic_coefficient")
def pymbolic_coefficient(container, lfs, index):
    # TODO introduce a proper type for local function spaces!
    if isinstance(lfs, str):
        valuearg(lfs, dtype=NumpyType("str"))

    # If the LFS is not yet a pymbolic expression, make it one
    from pymbolic.primitives import Expression
    if not isinstance(lfs, Expression):
        lfs = Variable(lfs)

    if isinstance(index, str):
        index = Variable(index)

    return Call(CoefficientAccess(container), (lfs, index,))


def type_coefficientcontainer():
    return "X"


def name_jacobian(restriction1, restriction2):
    # Restrictions may only differ if NONE
    if (restriction1 == Restriction.NONE) or (restriction2 == Restriction.NONE):
        assert restriction1 == restriction2
    return restricted_name(restricted_name("jac", restriction1), restriction2)


def type_jacobian():
    return "J"


def name_residual(restriction):
    return restricted_name("r", restriction)


def type_residual():
    return "R"


def name_accumulation_variable(restrictions=None):
    ft = get_global_context_value("form_type")
    measure = get_global_context_value("integral_type")
    if ft == 'residual' or ft == 'jacobian_apply':
        if restrictions is None:
            if measure == "cell":
                restrictions = (Restriction.NONE,)
            else:
                restrictions = (Restriction.POSITIVE,)
        return name_residual(*restrictions)
    if ft == 'jacobian':
        if restrictions is None:
            if measure == "cell":
                restrictions = (Restriction.NONE, Restriction.NONE)
            else:
                restrictions = (Restriction.POSITIVE, Restriction.POSITIVE)
        return name_jacobian(*restrictions)
    assert False


def type_accumulation_variable():
    ft = get_global_context_value("form_type")
    if ft == 'residual' or ft == 'jacobian_apply':
        return type_residual()
    if ft == 'jacobian':
        return type_jacobian()
    assert False
